import json
from copy import deepcopy


from ratio1 import Logger, const
from ratio1.bc import DefaultBlockEngine


if __name__ == '__main__':
  
  _DATA = {
    "9": 9,
    "2": 2,
    "3": 3,
    "10": {
      "2": 2,
      "100": 100,
      "1": 1
    },
    
  }
  
  PYTHON_MESSAGE = {
    **_DATA,
    "EE_SIGN": "MEQCIEIz_Nfy9CJ0GYW1V7Iw0uFJAVzu1TnOWkCVYnrt8PNHAiB0JCk_pgzGGIMz-KIvOCC_BzbGB8jxkAb_OwPX7AQTyA==",
    "EE_SENDER": "0xai_AuN2SENcYNzRgbPUHVCFe6W1q-vieUKap2VY9mU_Fljy",
    "EE_HASH": "7d72cf5bd6cda16c86dfb2c2c4983464edda5bf78e1ca3a21139f5037454c6f3"
  }
   
  
  l = Logger("ENC", base_folder=".", app_folder="_local_cache")
  eng1 = DefaultBlockEngine(
    log=l, 
    name="default",
    config={}
  )

  eng2_noneth = DefaultBlockEngine(
    eth_enabled=False,
    log=l, name="test1", 
    config={
        "PEM_FILE"     : "test1.pem",
        "PASSWORD"     : None,      
        "PEM_,LOCATION" : "data"
      }    
  )

  eng3_eth = DefaultBlockEngine(
    eth_enabled=True,
    log=l, name="test2", 
    config={
        "PEM_FILE"     : "test2.pem",
        "PASSWORD"     : None,      
        "PEM_,LOCATION" : "data"
      }    
  )
  
  v1 = eng1.verify(PYTHON_MESSAGE)
  l.P(f"check 1: {v1}", color='r' if not v1.valid else 'g')
  
  data = deepcopy(_DATA)
  eng1.sign(data)
  l.P(f"data: {data}")
  
  v2 = eng2_noneth.verify(data)
  l.P(f"check 2.1: {v2}", color='r' if not v2.valid else 'g')
  eng2_noneth.set_eth_flag(True)
  v2 = eng2_noneth.verify(data)
  l.P(f"check 2.2: {v2}", color='r' if not v2.valid else 'g')
  eng2_noneth.set_eth_flag(False)
  
  v3 = eng3_eth.verify(data)
  l.P(f"check 3: {v3}", color='r' if not v3.valid else 'g')  
  
  
  data_noneth = deepcopy(_DATA)
  eng2_noneth.sign(data_noneth)
  l.P(f"data: {data_noneth}")
  
  v4 = eng1.verify(data_noneth)
  l.P(f"check 4: {v4}", color='r' if not v4.valid else 'g')
  
  v5 = eng3_eth.verify(data_noneth)
  l.P(f"check 5: {v5}", color='r' if not v5.valid else 'g')


"""
ex10_telegram_echo_bot.py
-------------------------

A simple echo bot that replies to the user with the same message it received.
It also keeps track of the number of messages received from each user and the number of times
the processor has been called.

This example leverages the SDK ability to package user defined code and send them as handlers to
the target pipeline. 

"""
import os
import time

from ratio1 import Session, CustomPluginTemplate, PLUGIN_TYPES


def processor(plugin: CustomPluginTemplate):
  """
  This method will be continously called by the plugin to do any kind of required processing
  without being triggered by a message. This is useful for example to check for new events in
  a particular source or to do some kind of processing that is not directly related to a message
  """
  some_value = plugin.int_cache["app_counter"]
  some_value += 1
  if some_value > 1_000_000:
    # arbitrary condition to reset the value
    some_value = 0
  plugin.int_cache["app_counter"] = some_value
  return

def reply(plugin: CustomPluginTemplate, message: str, user: str):
  """
  This function is used to reply to a message. The given parameters are mandatory
  """
  # for each user message we increase a counter
  plugin.int_cache[user] += 1 # int_cache is a default dict that allows persistence in the plugin
  processor_value = plugin.int_cache["app_counter"]
  plugin.P(f"Replying to the {plugin.int_cache[user]} msg of '{user}' on message '{message}'")
  result = f"Processor at {processor_value}. The answer to your {plugin.int_cache[user]} question is in the question itself: '{message}'"
  return result


if __name__ == "__main__":   
  # Note: in order correctly setup the deployment you have to create the Session
  #       object that will configure the execution environment
  session = Session() 

  # NOTE: When working with SDK please use the nodes internal addresses. While the EVM address of the node
  #       is basically based on the same sk/pk it is in a different format and not directly usable with the SDK
  #       the internal node address is easily spoted as starting with 0xai_ and can be found 
  #       via `docker exec r1node get_node_info` or via the launcher UI

  # this tutorial assumes you have started your own local node for dev-testing purposes
  # you can either supply the node address via env or directly here
  my_node = os.getenv("EE_TARGET_NODE", "0xai_my_own_node_address") 

  session.wait_for_node(my_node) 
    
      
  # now we create a telegram bot pipeline & plugin instance and for that we only need the Telegram token
  # we can chose to use the token directly via `telegram_bot_token` parameter
  # or use the environment key EE_TELEGRAM_BOT_TOKEN 
  # in this case for this simple example we are going to use the token directly
  pipeline, _ = session.create_telegram_simple_bot(
    node=my_node,
    name="telegram_bot_echo",
    # telegram_bot_token="your_token_goes_here",  # we use the token directly
    message_handler=reply,
    processor_handler=processor,
  )
  
  pipeline.deploy() # we deploy the pipeline

  # Observation:
  #   next code is not mandatory - it is used to keep the session open and cleanup the resources
  #   due to the fact that this is a example/tutorial and maybe we dont want to keep the pipeline
  #   active after the session is closed we use close_pipelines=True
  #   in production, you would not need this code as the script can close 
  #   after the pipeline will be sent 
  session.wait(
    seconds=60,            # we wait the session for 120 seconds
    close_pipelines=True,   # we close the pipelines after the session
    close_session=True,     # we close the session after the session
  )



import traceback
import numpy as np

from collections import OrderedDict, deque
from time import perf_counter, sleep, time


DEFAULT_SECTION = 'main'
DEFAULT_THRESHOLD_NO_SHOW = 0
PERIODS = [60 * 60, 24 * 60 * 60, 30 * 24 * 60 * 60]
PERIODS_STR = ["hour", "day", "month"]
MAX_PERIOD_LAPS = 20
DEFAULT_PERIODIC_MULTIPLIER = 3

MAX_LAPS = 100
ZERO_THRESHOLD = 5e-4

_OBSOLETE_SECTION_TIME = 3600  # sections older than 1 hour are archived

class _TimersMixin(object):
  """
  Mixin for timers functionalities that are attached to `ratio1.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `ratio1.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_TimersMixin, self).__init__()
    self.timers = None
    self.timer_level = None # no actual purpose following addition of graph
    self.sections_last_used = {}
    self.opened_timers = None
    self.timers_graph = None
    self._timer_error = None
    self.default_timers_section = DEFAULT_SECTION
    self.__timer_mutex = False

    self.start_show_timer = None

    self.reset_timers()
    return

  def _maybe_create_timers_section(self, section=None):
    section = section or self.default_timers_section

    if section in self.timers:
      return

    self.timers[section] = OrderedDict()
    self.timer_level[section] = 0
    self.opened_timers[section] = deque()
    self.timers_graph[section] = OrderedDict()
    self.timers_graph[section]["ROOT"] = {"SLOW" : OrderedDict(), "FAST" : OrderedDict()}
    self._timer_error[section] = False
    return

  def reset_timers(self):
    self.timers = {}
    self.timer_level = {}
    self.opened_timers = {}
    self.timers_graph = {}
    self._timer_error = {}

    self._maybe_create_timers_section()
    return

  @staticmethod
  def get_empty_timer():
    return {
      'MEAN': 0,
      'MAX': 0,
      'COUNT': 0,
      'START': 0,
      'END': 0,
      'PASS': True,
      'LEVEL': 0,

      'START_COUNT': 0,
      'STOP_COUNT': 0,
      
      'LAPS' : deque(maxlen=MAX_LAPS),
      'HISTORY': {
        'LAST': [None for _ in range(len(PERIODS))],
        'DEQUES': [deque(maxlen=MAX_PERIOD_LAPS) for _ in range(len(PERIODS))],
      }
    }

  def restart_timer(self, sname, section=None):
    section = section or self.default_timers_section
    self.timers[section][sname] = self.get_empty_timer()
    return

  def _add_in_timers_graph(self, sname, section=None):
    section = section or self.default_timers_section
    self.timers[section][sname]['LEVEL'] = self.timer_level[section]
    if sname not in self.timers_graph[section]:
      self.timers_graph[section][sname] = {"SLOW" : OrderedDict(), "FAST" : OrderedDict()}

    self.timers_graph[section][sname]["FAST"] = OrderedDict() ## there is no ordered set, so we use OrderedDict with no values
    return

  def start_timer(self, sname, section=None):
    section = section or self.default_timers_section
    if section == self.default_timers_section:
      assert self.is_main_thread, "Attempted to run threaded timer '{}' without section".format(sname)

    self._maybe_create_timers_section(section)

    if not self.DEBUG:
      return -1

    if sname not in self.timers[section]:
      self.restart_timer(sname, section)

    curr_time = perf_counter()
    self._add_in_timers_graph(sname, section=section)
    self.timers[section][sname]['START'] = curr_time
    self.timers[section][sname]['START_COUNT'] += 1
    if len(self.opened_timers[section]) >= 1:
      parent = self.opened_timers[section][-1]
    else:
      parent = "ROOT"
    #endif

    self.timers_graph[section][parent]["FAST"][sname] = None
    self.timers_graph[section][parent]["SLOW"][sname] = None

    self.timer_level[section] += 1
    self.opened_timers[section].append(sname)

    faulty_timers = self._get_section_faulty_timers(section)
    if len(faulty_timers) > 0 and not self._timer_error[section]:
      self.P("Something is wrong with the timers in section '{}':".format(section), color='r')
      for ft in faulty_timers:
        self.P("  {}: {}".format(ft, self.timers[section][ft]), color='r')
      self._timer_error[section] = True
    #endif

    return curr_time

  def get_time_until_now(self, sname, section=None):
    section = section or self.default_timers_section
    ctimer = self.timers[section][sname]
    return perf_counter() - ctimer['START']

  def get_faulty_timers(self):
    dct_faulty = {}
    for section in self.timers:
      dct_faulty[section] = self._get_section_faulty_timers(section)
    return dct_faulty

  def _get_section_faulty_timers(self, section=None):
    section = section or self.default_timers_section
    lst_faulty = []
    for tmr_name, tmr in self.timers[section].items():
      if (tmr['START_COUNT'] - tmr['STOP_COUNT']) > 1:
        lst_faulty.append(tmr_name)
    return lst_faulty

  def end_timer_no_skip(self, sname, section=None, periodic=False):
    return self.end_timer(sname, skip_first_timing=False, section=section, periodic=periodic)

  def get_opened_timer(self, section=None):
    section = section or self.default_timers_section
    timer = self.opened_timers[section][-1]
    return timer, perf_counter() - self.timers[section][timer]['START']

  def get_periodic_multiplier(self):
    if self.config_data is None:
      return DEFAULT_PERIODIC_MULTIPLIER
    return self.config_data.get('PERIODIC_MULTIPLIER', DEFAULT_PERIODIC_MULTIPLIER)

  def add_periodic_record(self, sname, record, section=None, idx=0, check=False):
    chistory = self.timers[section][sname]['HISTORY']
    dq_size = len(chistory['DEQUES'][idx])
    if check and dq_size > 2:
      cmean = np.array(chistory['DEQUES'][idx]).mean()
      cstd = np.std(chistory['DEQUES'][idx])
      cnt = self.timers[section][sname]['COUNT']
      q = self.get_periodic_multiplier()
      limit = cmean + cstd * q
      if limit < record:
        self.P(
          f"[WARNING]Timer {'' if section is None else section + '|'}{sname} "
          f"took longer than expected at iter {cnt} based on {dq_size} records ("
          f"{round(record, 3)} > {round(limit, 3)}[{round(cmean, 3)}+{round(q, 3)}*{round(cstd, 3)}]) [{PERIODS_STR[idx]}]",
          color='r'
        )
      # endif warning
    # endif anomaly check
    chistory['DEQUES'][idx].append(record)
    chistory['LAST'][idx] = self.timers[section][sname]['END']
    return

  def end_timer(self, sname, skip_first_timing=False, section=None, periodic=False):
    section = section or self.default_timers_section
    if sname not in self.timers[section]:
      return
    result = 0
    self.sections_last_used[section] = time()
    if self.DEBUG:
      self.opened_timers[section].pop()
      self.timer_level[section] -= 1

      ctimer = self.timers[section][sname]
      ctimer['STOP_COUNT'] += 1
      ctimer['END'] = perf_counter()
      result = ctimer['END'] - ctimer['START']
      ctimer['LAPS'].append(result)
      if periodic:
        chistory = ctimer['HISTORY']
        for i, period in enumerate(PERIODS):
          if chistory['LAST'][i] is None:
            self.add_periodic_record(sname, result, section=section, idx=i)
          elif ctimer['END'] - chistory['LAST'][i] > period:
            self.add_periodic_record(sname, result, section=section, idx=i, check=True)
          # endif periodic record
        # endfor i, period
      # endif periodic
      _count = ctimer['COUNT']
      _prev_avg = ctimer['MEAN']
      avg = _count * _prev_avg

      if ctimer['PASS'] and skip_first_timing:
        ctimer['PASS'] = False
        return result  # do not record first timing in average nor the max

      ctimer['MAX'] = max(ctimer['MAX'], result)

      ctimer['COUNT'] = _count + 1
      avg += result
      avg = avg / ctimer["COUNT"]
      ctimer['MEAN'] = avg
    return result

  def stop_timer(self, sname, skip_first_timing=False, section=None, periodic=False):
    return self.end_timer(sname=sname, skip_first_timing=skip_first_timing, section=section, periodic=periodic)

  def show_timer_total(self, sname, section=None):
    section = section or self.default_timers_section
    ctimer = self.timers[section][sname]
    cnt = ctimer['COUNT']
    val = ctimer['MEAN'] * cnt
    self.P("  {} = {:.3f} in {} laps".format(sname, val, cnt))
    return

  def _format_timer(self, key, section, was_recently_seen,
                   summary='mean',
                   show_levels=True,
                   show_max=True,
                   show_last=True,
                   show_count=True,
                   div=None,
                   threshold_no_show=None,
                   max_key_size=30,
                   ):

    if threshold_no_show is None:
      threshold_no_show = DEFAULT_THRESHOLD_NO_SHOW

    ctimer = self.timers[section].get(key, None)

    if ctimer is None:
      return

    mean_time = ctimer['MEAN']
    np_laps = np.array(ctimer['LAPS'])
    if len(np_laps) > 0:
      laps_mean = np_laps.mean()
      laps_std = np_laps.std()
      laps_zcount = (np_laps <= ZERO_THRESHOLD).sum()
      laps_nzcount = len(np_laps) - laps_zcount
      laps_nz_mean = np_laps.sum() / laps_nzcount if laps_nzcount > 0 else -1
      laps_low_cnt = (np_laps <= max(ZERO_THRESHOLD, laps_mean - laps_std)).sum()
      laps_low_prc =  laps_low_cnt / np_laps.shape[0] * 100
    else:
      # not mandatory but nice for forever opened 1 time timers
      laps_mean = -1
      laps_std = -1
      laps_zcount = -1
      laps_nzcount = -1
      laps_nz_mean = -1
      laps_low_cnt = -1
      laps_low_prc =  -1
      # end not mandatory

    count = ctimer['COUNT']
    
    if mean_time < threshold_no_show:
      return

    max_time = ctimer['MAX']
    current_time = np_laps[-1] if len(np_laps) > 0 else -1 # ctimer['END'] - ctimer['START']

    if not was_recently_seen:
      key = '[' + key[:max_key_size] + ']'

    if show_levels:
      s_key = '  ' * ctimer['LEVEL'] + key[:max_key_size]
    else:
      s_key = key[:max_key_size]
    msg = None
    if summary in ['mean', 'avg']:
      # self.verbose_log(" {} = {:.4f}s (max lap = {:.4f}s)".format(s_key,mean_time,max_time))
      msg = " {} = {:.4f}s/q:{:.4f}s/nz:{:.4f}s".format(s_key, mean_time, laps_mean, laps_nz_mean)
    else:
      # self.verbose_log(" {} = {:.4f}s (max lap = {:.4f}s)".format(s_key,total, max_time))
      total = mean_time * count
      msg = " {} = {:.4f}s".format(s_key, total)
    if show_max:
      msg += ", max: {:.4f}s".format(max_time)
    if show_last:
      msg += ", lst: {:.4f}s".format(current_time)
    if show_count:
      msg += ", c: {}/L:{:.0f}%".format(count, laps_low_prc)
    if div is not None:
      msg += ", itr(B{}): {:.4f}s".format(div, mean_time / div)
    return msg

  def show_timers(self, indent=4, **kwargs):
    color = kwargs.pop('color', 'n')
    lst_logs = self.format_timers(**kwargs)
    if indent > 0:
      lst_logs = [" " * indent + l for l in lst_logs]
    if len(lst_logs) > 0:
      full_log = "\n".join(lst_logs)      
      self.verbose_log(full_log.lstrip(), color=color)
    return

  def show_timings(self, **kwargs):
    self.show_timers(**kwargs)
    return

  def format_timers(self, summary=None,
                  title=None,
                  show_levels=True,
                  show_max=True,
                  show_last=True,
                  show_count=True,
                  div=None,
                  threshold_no_show=None,
                  selected_sections=None,
                  obsolete_section_time=_OBSOLETE_SECTION_TIME,
                  ):

    if selected_sections is not None:
      assert isinstance(selected_sections, list)

    if threshold_no_show is None:
      threshold_no_show = DEFAULT_THRESHOLD_NO_SHOW

    if summary is None:
      summary = 'mean'

    if title is None:
      title = ''
    
    lst_logs = []
    self.start_show_timer = time()
    try:
      while self.__timer_mutex:
        elapsed_in_show = time() - self.start_show_timer
        if elapsed_in_show > 15:
          raise ValueError("show_timers: Run time exceeded! Mutex is locked for more than 15 seconds.")
        sleep(0.001)
      # end while
      self.__timer_mutex = True
      
      self.start_show_timer = time()
      self.__dfs_stack = deque(maxlen=100)
      
      def dfs(visited, graph, node, was_recently_seen, logs, sect):
        self.__dfs_stack.append(node)
        elapsed_in_show = time() - self.start_show_timer
        
        if elapsed_in_show > 5:
          raise ValueError("show_timers: Run time exceeded! DFS took longer than 5 seconds. Stack: {}".format(self.__dfs_stack))
        
        if node not in visited:
          formatted_node = self._format_timer(
            key=node,
            section=sect,
            was_recently_seen=was_recently_seen,
            summary=summary,
            show_levels=show_levels, show_last=show_last,
            show_max=show_max, show_count=show_count, div=div,
            threshold_no_show=threshold_no_show
          )
          if formatted_node is not None:
            logs.append(formatted_node)
          visited.add(node)
          slow_keys = list(graph[node]["SLOW"].keys())
          fast_keys = list(graph[node]["FAST"].keys())
          for neighbour in slow_keys:
            recently_seen = neighbour in fast_keys
            dfs(visited, graph, neighbour, recently_seen, logs, sect)
          #endfor
        #endif
      #enddef

      if self.DEBUG:
        if len(title) > 0:
          title = ' ' + title
        header = "Timing results{} at {}:".format(title, self.now_str(nice_print=True, short=True))
        if threshold_no_show > 0:
          header += " (discarding entries with time < {})".format(threshold_no_show)
        lst_logs.append(header)

        ## SORTING sections and keeping the default section the first one ..
        keys = list(self.timers.keys())
        if selected_sections is not None:
          keys = selected_sections

        add_back_default_section = False
        if self.default_timers_section in keys:
          add_back_default_section = True
          keys.pop(keys.index(self.default_timers_section))
        keys.sort()
        if add_back_default_section:
          keys = [self.default_timers_section] + keys

        old_sections = []
        for section in keys:
          last_see_ago = None
          if section in self.sections_last_used:
            last_see_ago = time() - self.sections_last_used[section]
            if last_see_ago > obsolete_section_time:
              old_sections.append(section)
              continue
          lst_logs.append("Section '{}'{}".format(
            section, " last seen {:.1f}s ago".format(last_see_ago) if last_see_ago is not None else ""
          ))
          buffer_visited = set()
          dfs(buffer_visited, self.timers_graph[section], "ROOT", True, lst_logs, section)
        if len(old_sections) > 0:
          lst_logs.append("Archived {} sections older than {:.1f} hrs.".format(
            len(old_sections), obsolete_section_time / 3600, 
          ))
      else:
        self.verbose_log("DEBUG not activated!")
    except Exception as exc:
      self.P("EXCEPTION in show_timers: {}. Stacktrace:\n{}".format(exc, traceback.format_exc()), color='r')
    self.__timer_mutex = False
    return lst_logs

  def get_timing_dict(self, skey, section=None):
    section = section or self.default_timers_section
    timers_section = self.timers.get(section, {})
    dct = timers_section.get(skey, {})
    return dct

  def get_timer(self, skey, section=None):
    return self.get_timing_dict(skey, section=section)

  def get_timer_overall_mean(self, skey, section=None):
    tmr = self.get_timer(skey, section=section)
    result = tmr.get('MEAN', 0)
    return result

  def get_timer_mean(self, skey, section=None):
    tmr = self.get_timer(skey, section=section)
    laps = tmr.get('LAPS', [])
    result = np.mean(laps) if len(laps) > 0 else -1
    return result


  def get_timer_count(self, skey, section=None):
    tmr = self.get_timer(skey, section=section)
    result = tmr.get('COUNT', 0)
    return result
  
  def import_timers_section(self, dct_timers, dct_timers_graph, section, overwrite=False):
    if self.__timer_mutex:
      # we skip this import until next time
      self.P("WARNING: Cannot import section '{}' with {} timers while processing sections!".format(
        section, len(dct_timers)
      ), color='r')
      return
    if section in self.timers and not overwrite:
      self.P("WARNING: Cannot import section '{}' with {} timers as there is already a existing section".format(
        section, len(dct_timers)
      ), color='r')
      return False
    self.timers[section] = dct_timers
    self.timers_graph[section] = dct_timers_graph
    self.sections_last_used[section] = time()
    return True
  
  def export_timers_section(self, section=None):
    section = section or self.default_timers_section
    if section not in self.timers:
      self.P("WARNING: Cannot export unexisting timers section '{}'".format(
        section
      ), color='r')
      return None, None
    dct_timers = self.timers[section]
    dct_timers_graph = self.timers_graph[section]
    return dct_timers, dct_timers_graph



import os
import base64
import json
import binascii
import numpy as np
import datetime
import uuid
import requests

from collections import defaultdict
from hashlib import sha256, md5
from threading import Lock
from copy import deepcopy


from cryptography.hazmat.primitives import serialization

from ..utils.config import get_user_folder


from ..const.base import (
  BCctbase, BCct, 
  DAUTH_SUBKEY, DAUTH_ENV_KEY,
  DAUTH_NONCE, dAuth,
)

from .evm import _EVMMixin, Web3, EE_VPN_IMPL

EVM_COMMENT = " # "
INVALID_COMMENT = " # INVALID: "
  
  
class _DotDict(dict):
  __getattr__ = dict.__getitem__
  __setattr__ = dict.__setitem__
  __delattr__ = dict.__delitem__
  
  
class VerifyMessage(_DotDict):
  def __init__(self):
    self.valid = False
    self.message = None
    self.sender = None
    
    
ALL_NON_DATA_FIELDS = [val for key, val in BCctbase.__dict__.items() if key[0] != '_']

NO_ETH_NON_DATA_FIELDS = [
  val for key, val in BCctbase.__dict__.items() 
  if key[0] != '_' and not key.startswith('ETH_')
]

def replace_nan_inf(data, inplace=False):
  assert isinstance(data, (dict, list)), "Only dictionaries and lists are supported"
  if inplace:
    d = data
  else:
    d = deepcopy(data)    
  stack = [d]
  while stack:
    current = stack.pop()
    for key, value in current.items():
      if isinstance(value, dict):
        stack.append(value)
      elif isinstance(value, list):
        for item in value:
          if isinstance(item, dict):
            stack.append(item)
      elif isinstance(value, float) and (np.isnan(value) or np.isinf(value)):
        current[key] = None
  return d 

class _SimpleJsonEncoder(json.JSONEncoder):
  """
  Used to help jsonify numpy arrays or lists that contain numpy data types.
  """
  def default(self, obj):
    if isinstance(obj, np.integer):
      return int(obj)
    elif isinstance(obj, np.floating):
      return float(obj)
    elif isinstance(obj, np.ndarray):
      return obj.tolist()
    elif isinstance(obj, np.ndarray):
      return obj.tolist()
    elif isinstance(obj, datetime.datetime):
      return obj.strftime("%Y-%m-%d %H:%M:%S")
    else:
      return super(_SimpleJsonEncoder, self).default(obj)

class _ComplexJsonEncoder(json.JSONEncoder):
  def default(self, obj):
    if isinstance(obj, np.integer):
      return int(obj)
    elif isinstance(obj, np.floating):
      return float(obj)
    elif isinstance(obj, np.ndarray):
      return obj.tolist()
    elif isinstance(obj, datetime.datetime):
      return obj.strftime("%Y-%m-%d %H:%M:%S")
    else:
      return super(_ComplexJsonEncoder, self).default(obj)

  def iterencode(self, o, _one_shot=False):
    """Encode the given object and yield each string representation as available."""
    if self.check_circular:
      markers = {}
    else:
      markers = None
    if self.ensure_ascii:
      _encoder = json.encoder.encode_basestring_ascii
    else:
      _encoder = json.encoder.encode_basestring
    
    def floatstr(o, allow_nan=self.allow_nan, _repr=float.__repr__, _inf=json.encoder.INFINITY, _neginf=-json.encoder.INFINITY):
      if o != o:  # Check for NaN
        text = 'null'
      elif o == _inf:
        text = 'null'
      elif o == _neginf:
        text = 'null'
      else:
        return repr(o).rstrip('0').rstrip('.') if '.' in repr(o) else repr(o)

      if not allow_nan:
        raise ValueError("Out of range float values are not JSON compliant: " + repr(o))
      
      return text

    _iterencode = json.encoder._make_iterencode(
      markers, self.default, _encoder, self.indent, floatstr,
      self.key_separator, self.item_separator, self.sort_keys,
      self.skipkeys, _one_shot
    )
    return _iterencode(o, 0)

## RIPEMD160

# Message schedule indexes for the left path.
ML = [
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
    7, 4, 13, 1, 10, 6, 15, 3, 12, 0, 9, 5, 2, 14, 11, 8,
    3, 10, 14, 4, 9, 15, 8, 1, 2, 7, 0, 6, 13, 11, 5, 12,
    1, 9, 11, 10, 0, 8, 12, 4, 13, 3, 7, 15, 14, 5, 6, 2,
    4, 0, 5, 9, 7, 12, 2, 10, 14, 1, 3, 8, 11, 6, 15, 13
]

# Message schedule indexes for the right path.
MR = [
    5, 14, 7, 0, 9, 2, 11, 4, 13, 6, 15, 8, 1, 10, 3, 12,
    6, 11, 3, 7, 0, 13, 5, 10, 14, 15, 8, 12, 4, 9, 1, 2,
    15, 5, 1, 3, 7, 14, 6, 9, 11, 8, 12, 2, 10, 0, 4, 13,
    8, 6, 4, 1, 3, 11, 15, 0, 5, 12, 2, 13, 9, 7, 10, 14,
    12, 15, 10, 4, 1, 5, 8, 7, 6, 2, 13, 14, 0, 3, 9, 11
]

# Rotation counts for the left path.
RL = [
    11, 14, 15, 12, 5, 8, 7, 9, 11, 13, 14, 15, 6, 7, 9, 8,
    7, 6, 8, 13, 11, 9, 7, 15, 7, 12, 15, 9, 11, 7, 13, 12,
    11, 13, 6, 7, 14, 9, 13, 15, 14, 8, 13, 6, 5, 12, 7, 5,
    11, 12, 14, 15, 14, 15, 9, 8, 9, 14, 5, 6, 8, 6, 5, 12,
    9, 15, 5, 11, 6, 8, 13, 12, 5, 12, 13, 14, 11, 8, 5, 6
]

# Rotation counts for the right path.
RR = [
    8, 9, 9, 11, 13, 15, 15, 5, 7, 7, 8, 11, 14, 14, 12, 6,
    9, 13, 15, 7, 12, 8, 9, 11, 7, 7, 12, 7, 6, 15, 13, 11,
    9, 7, 15, 11, 8, 6, 6, 14, 12, 13, 5, 14, 13, 13, 7, 5,
    15, 5, 8, 11, 14, 14, 6, 14, 6, 9, 12, 9, 12, 5, 15, 8,
    8, 5, 12, 9, 12, 5, 14, 6, 8, 13, 6, 5, 15, 13, 11, 11
]

# K constants for the left path.
KL = [0, 0x5a827999, 0x6ed9eba1, 0x8f1bbcdc, 0xa953fd4e]

# K constants for the right path.
KR = [0x50a28be6, 0x5c4dd124, 0x6d703ef3, 0x7a6d76e9, 0]


def fi(x, y, z, i):
    """The f1, f2, f3, f4, and f5 functions from the specification."""
    if i == 0:
        return x ^ y ^ z
    elif i == 1:
        return (x & y) | (~x & z)
    elif i == 2:
        return (x | ~y) ^ z
    elif i == 3:
        return (x & z) | (y & ~z)
    elif i == 4:
        return x ^ (y | ~z)
    else:
        assert False


def rol(x, i):
    """Rotate the bottom 32 bits of x left by i bits."""
    return ((x << i) | ((x & 0xffffffff) >> (32 - i))) & 0xffffffff


def compress(h0, h1, h2, h3, h4, block):
    """Compress state (h0, h1, h2, h3, h4) with block."""
    # Left path variables.
    al, bl, cl, dl, el = h0, h1, h2, h3, h4
    # Right path variables.
    ar, br, cr, dr, er = h0, h1, h2, h3, h4
    # Message variables.
    x = [int.from_bytes(block[4*i:4*(i+1)], 'little') for i in range(16)]

    # Iterate over the 80 rounds of the compression.
    for j in range(80):
        rnd = j >> 4
        # Perform left side of the transformation.
        al = rol(al + fi(bl, cl, dl, rnd) + x[ML[j]] + KL[rnd], RL[j]) + el
        al, bl, cl, dl, el = el, al, bl, rol(cl, 10), dl
        # Perform right side of the transformation.
        ar = rol(ar + fi(br, cr, dr, 4 - rnd) + x[MR[j]] + KR[rnd], RR[j]) + er
        ar, br, cr, dr, er = er, ar, br, rol(cr, 10), dr

    # Compose old state, left transform, and right transform into new state.
    return h1 + cl + dr, h2 + dl + er, h3 + el + ar, h4 + al + br, h0 + bl + cr


def ripemd160(data):
    """Compute the RIPEMD-160 hash of data."""
    # Initialize state.
    state = (0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476, 0xc3d2e1f0)
    # Process full 64-byte blocks in the input.
    for b in range(len(data) >> 6):
        state = compress(*state, data[64*b:64*(b+1)])
    # Construct final blocks (with padding and size).
    pad = b"\x80" + b"\x00" * ((119 - len(data)) & 63)
    fin = data[len(data) & ~63:] + pad + (8 * len(data)).to_bytes(8, 'little')
    # Process final blocks.
    for b in range(len(fin) >> 6):
        state = compress(*state, fin[64*b:64*(b+1)])
    # Produce output.
    return b"".join((h & 0xffffffff).to_bytes(4, 'little') for h in state)
  
# END ## RIPEMD160  

class BaseBlockEngine(_EVMMixin):
  """
  This multiton (multi-singleton via key) is the base workhorse of the private blockchain. 
  
  Parameters
  ----------
  
  name: str
    the name of the engine. Used to create the private key file name.
    
  config: dict
    the configuration dict that contains the PEM_FILE, PASSWORD, PEM_LOCATION keys
    for configuring the private key file access
    
  log: Logger object
    the Logger object
    
  ensure_ascii_payloads: bool
    flag that controls if the payloads are encoded as ascii or not. Default `False` for JS compatibility.

  user_config: bool
    flag that controls if the keys are stored in the user private folder or in the data folder of the _local_cache
  
  """
  _lock: Lock = Lock()
  _whitelist_lock : Lock = Lock()
  __instances = {}
  
  def __new__(
    cls, 
    name, 
    log, 
    config={}, 
    ensure_ascii_payloads=False, 
    verbosity=1, 
    user_config=False,   
    eth_enabled=True, 
  ):
    with cls._lock:
      if name not in cls.__instances:
        instance = super(BaseBlockEngine, cls).__new__(cls)
        instance._build(
          name=name, log=log, config=config, 
          ensure_ascii_payloads=ensure_ascii_payloads,
          verbosity=verbosity,
          user_config=user_config,
          eth_enabled=eth_enabled,
        )
        cls.__instances[name] = instance
      else:
        instance = cls.__instances[name]
    return instance
  
  def _build(
      self, 
      name, 
      config:dict, 
      log=None, 
      ensure_ascii_payloads=False,
      verbosity=1,
      user_config=False,
      eth_enabled=True,
    ):

    self.__name = name
    assert log is not None, "Logger object was not provided!"      
      
    self.log = log
    
    self._first_checks_done = defaultdict(lambda: False) # used to store the first check results
    
    self.__private_key = None
    self.__verbosity = verbosity
    self.__public_key = None    
    self.__password = config.get(BCct.K_PASSWORD)    
    self.__config = config
    self.__ensure_ascii_payloads = ensure_ascii_payloads
    
    self._eth_enabled = eth_enabled
    
    if user_config:
      user_folder = get_user_folder()
      pem_fn = str(user_folder / BCct.USER_PEM_FILE)
      self.P(f"Home config selected. Setting pem_fn to {pem_fn}")
    else:
      pem_name = config.get(BCct.K_PEM_FILE, BCct.DEFAULT_PEM_FILE)
      pem_folder = config.get(BCct.K_PEM_LOCATION, BCct.DEFAULT_PEM_LOCATION)
      pem_fn = os.path.join(log.get_target_folder(pem_folder), pem_name)
      self.P(f"Arbitrary config selected. Setting pem_fn to {pem_fn}")
    #endif pem is defined in ~/.ratio1/ or in the data folder of the _local_cache
    self.__pem_file = pem_fn
    self._init()
    return


  def P(self, s, color=None, boxed=False, verbosity=1, **kwargs):
    if verbosity > self.__verbosity:
      return
    if not boxed:
      s = "<BC:{}> ".format(self.__name) + s
    return self.log.P(
      s, 
      color=color or 'd', 
      boxed=boxed, 
      **kwargs
    )
    return
  
  
  @property
  def eth_enabled(self):
    return self._eth_enabled


  def set_eth_flag(self, value):    
    if value != self._eth_enabled:
      self._eth_enabled = value
      self.log.P("Changed eth_enabled to {}".format(value), color='d')
    return


  @property
  def name(self):
    return self.__name

  
  @property
  def eth_address(self):
    return self.__eth_address


  @property
  def eth_account(self):
    return self.__eth_account


  def _init(self):
    self.P(
      f"Initializing BC-engine (ETH_ENABLED={self._eth_enabled})...", verbosity=1
    )

    if True:
      self.P("Initializing private blockchain:\n{}".format(
        json.dumps(self.__config, indent=4)), verbosity=1
      )
    if self.__pem_file is not None:
      try:
        full_path = os.path.abspath(self.__pem_file)
        self.P("Trying to load sk from {}".format(full_path), verbosity=1)
        self.__private_key = self._text_to_sk(
          source=self.__pem_file,
          from_file=True,
          password=self.__password,
        )
        self.P("  Loaded sk from {}".format(full_path), verbosity=1)        
      except:
        self.P("  Failed to load sk from {}".format(full_path), color='r', verbosity=1)

    if self.__private_key is None:
      self.P("Creating new private key", verbosity=1)
      self.__private_key = self._create_new_sk()
      self._sk_to_text(
        private_key=self.__private_key,
        password=self.__password,
        fn=self.__pem_file,
      )
    
    os.environ[BCct.K_USER_CONFIG_PEM_FILE] = os.path.abspath(self.__pem_file)
    
    self.__public_key = self._get_pk(private_key=self.__private_key)    
    self.__address = self._pk_to_address(self.__public_key)
    ### Ethereum
    self.__eth_address = self._get_eth_address()
    self.__eth_account = self._get_eth_account()
    ### end Ethereum
    if self._eth_enabled:
      self.P(
        "{} / ETH: {} ({})".format(self.address, self.eth_address, self.evm_network), boxed=True, verbosity=1,
        color='g'
      )
    else:
      self.P("Address: {}".format(self.address), boxed=True, color='g', verbosity=1)
    self.P("Allowed list of senders: {}".format(self.allowed_list), verbosity=1)
    return
  
  @property
  def private_key(self):
    return self.__private_key
  
  
  @property
  def public_key(self):
    return self.private_key.public_key()
  
 
  @staticmethod
  def _compute_hash(data : bytes, method='SHA256'):
    """
    Computes the hash of a `bytes` data message

    Parameters
    ----------
    data : bytes
      the input message usually obtained from a bynary jsoned dict.
      
    method : str, optional
      the hash algo. The default is 'HASH160'.


    Returns
    -------
    result : bytes, str
      hash both in bin and text format.

    """
    result = None, None
    method = method.upper()
    assert method in ['HASH160', 'SHA256', 'MD5']
        
    if method == 'MD5':
      hash_obj = md5(data)
      result = hash_obj.digest(), hash_obj.hexdigest()
    elif method == 'SHA256':
      hash_obj = sha256(data)
      result = hash_obj.digest(), hash_obj.hexdigest()
    elif method == 'HASH160':
      hb_sha256 = sha256(data).digest()
      hb_h160 = ripemd160(hb_sha256)
      result = hb_h160, binascii.hexlify(hb_h160).decode()
    return result  
  
  
  @staticmethod
  def _binary_to_text(data : bytes, method='base64'):
    """
    Encodes a bytes message as text

    Parameters
    ----------
    data : bytes
      the binary data, usually a signature, hash, etc.
      
    method : str, optional
      the method - 'base64' or other. The default is 'base64'.


    Returns
    -------
    txt : str
      the base64 or hexlified text.

    """
    assert isinstance(data, bytes)
    if method == 'base64':
      txt = base64.urlsafe_b64encode(data).decode()
    else:
      txt = binascii.hexlify(data).decode()
    return txt
  
  
  @staticmethod
  def _text_to_binary(text : str, method='base64'):
    """
    Convert from str/text to binary

    Parameters
    ----------
    text : str
      the message.
      
    method : str, optional
      the conversion method. The default is 'base64'.


    Returns
    -------
    data : bytes
      the decoded binary message.

    """
    assert isinstance(text, str), "Cannot convert non text to binary"
    if method == 'base64':
      data = base64.urlsafe_b64decode(text)
    else:
      data = binascii.unhexlify(text)
    return data  

  
  @staticmethod
  def _get_pk(private_key):
    """
    Simple wrapper to generate pk from sk


    Returns
    -------
    public_key : pk
    
    """
    return private_key.public_key()
  
  
  def _get_allowed_file(self):
    """
    Return the file path for the autorized addresses
    """
    folder = self.log.base_folder
    path = os.path.join(folder, BCct.AUTHORISED_ADDRS)
    return path  
  
  
  def address_is_valid(self, address, return_error=False):
    """
    Checks if an address is valid

    Parameters
    ----------
    address : str
      the text address.

    Returns
    -------
    bool
      True if the address is valid.

    """
    result = False
    msg = ""
    try:
      pk = self._address_to_pk(address)
      result = False if pk is None else True
    except Exception as exc:
      result = False
      msg = str(exc)
    if return_error:
      return result, msg
    return result
  
  
  def add_address_to_allowed(self, address : any):
    """
    Adds a new address or a list of addresses to the allowed list
    """
    changed = False
    if isinstance(address, str):
      address = [address]
    #endif
    if isinstance(address, list) and len(address) > 0:
      # self.P(f"Adding addresses to the allowed list:\n{address}", verbosity=1)
      # now check addresses
      lst_lines = []
      lst_addrs = []
      lst_names = []
      whitelist = self.whitelist_with_prefixes
      for addr in address:
        addr = addr.strip()
        parts = addr.split()
        if len(parts) == 0:
          continue
        addr = parts[0]
        name = parts[1] if len(parts) > 1 else ""
        is_valid, valid_msg = self.address_is_valid(addr, return_error=True)
        if not is_valid:
          self.P("WARNING: address <{}> is not valid. Ignoring.".format(addr), color='r')
          addr = "# " + addr
          name = name + INVALID_COMMENT + valid_msg
          continue # skip invalid address or go forward and add it ...
        else:
          addr = self.maybe_add_prefix(addr)
          if addr in whitelist:
            self.P("WARNING: address <{}> already in the allowed list. Ignoring.".format(addr), color='r')
            continue
          eth = self.node_address_to_eth_address(addr)
          name = name + EVM_COMMENT + eth          
        str_line = "{}{}".format(addr, ("  " + name) if len(name)>0 else "")
        lst_lines.append(str_line)
        lst_addrs.append(addr)
        lst_names.append(name)
        #endif
      #endfor
      if len(lst_lines) > 0:
        with self._whitelist_lock:
          fn = self._get_allowed_file()
          with open(fn, 'rt') as fh:
            lst_existing = fh.readlines()
          #endwith
        for line, addr, name in zip(lst_lines, lst_addrs, lst_names):
          if line not in lst_existing:
            changed = True
            lst_existing.append(line)
            self.P("Address <{}> added to the allowed list.".format(addr), color='g')
        #endfor
        if changed:
          with self._whitelist_lock:
            fn = self._get_allowed_file()
            with open(fn, 'wt') as fh:
              for line in lst_existing:
                line = line.strip()
                if line != "":
                  fh.write(f"{line}\n")
              #endfor each address in modified whitelist
            #endwith open file
          #endwith lock
        #endif changed
      #endif addresses received ok
    return changed


  def _load_and_maybe_create_allowed(self, return_names=False, return_prefix=False):
    lst_final = []
    lst_names = []
    with self._whitelist_lock:
      try:
        fn = self._get_allowed_file()
        lst_allowed = []
        if os.path.isfile(fn):
          with open(fn, 'rt') as fh:
            lst_allowed = fh.readlines()
        else:
          full_path = os.path.abspath(fn)
          self.P("WARNING: no `{}` file found. Creating empty one.".format(full_path), verbosity=1)
          with open(fn, 'wt') as fh:
            fh.write('\n')
        lst_allowed = [x.strip() for x in lst_allowed]
        lst_allowed = [x for x in lst_allowed if x != '']
        lst_lines_to_write = []
        needs_rewrite = False
        for allowed_tuple in lst_allowed:
          parts = allowed_tuple.split()
          if len(parts) == 0:
            continue
          allowed = parts[0]
          if allowed.startswith("#"):
            # skip comments but keep them if we re-write the file
            lst_lines_to_write.append(allowed_tuple)
            continue
          allowed = self._remove_prefix(allowed)
          name = parts[1] if len(parts) > 1 else ""
          is_valid, valid_msg = self.address_is_valid(allowed, return_error=True)
          if not is_valid:
            self.P("WARNING: address <{}> is not valid. Commenting {} from allowed list.".format(
              allowed, allowed_tuple), color='r'
            )
            needs_rewrite = True
            error_line = "# " + allowed_tuple + INVALID_COMMENT + valid_msg
            lst_lines_to_write.append(error_line)
          else:
            if return_prefix:
              allowed = self.maybe_add_prefix(allowed)
            lst_final.append(allowed)
            lst_names.append(name)
            if len(parts) < 3:
              eth = self.node_address_to_eth_address(allowed)
              allowed_tuple = allowed_tuple + EVM_COMMENT + eth
              needs_rewrite = True 
            lst_lines_to_write.append(allowed_tuple)
              
        if needs_rewrite:
          with open(fn, 'wt') as fh:
            for line in lst_lines_to_write:
              fh.write(f"{line}\n")
      except Exception as exc:
        self.P(f"ERROR: failed to load the allowed list of addresses: {exc}", color='r')
      #endtry
    #endwith
    if return_names:
      return lst_final, lst_names
    return lst_final
  
        
  def _remove_prefix(self, address):
    """
    Removes the prefix from the address

    Parameters
    ----------
    address : str
      the text address.

    Returns
    -------
    address : str
      the address without the prefix.
    """
    if address.startswith(BCct.ADDR_PREFIX):
      address = address[len(BCct.ADDR_PREFIX):]
    elif address.startswith(BCct.ADDR_PREFIX_OLD):
      address = address[len(BCct.ADDR_PREFIX_OLD):]
    return address
  
  def _add_prefix(self, address):
    """
    Adds the prefix to the address

    Parameters
    ----------
    address : str
      the text address.

    Returns
    -------
    address : str
      the address with the prefix.
    """
    address = self._remove_prefix(address)
    address = BCct.ADDR_PREFIX + address  
    return address
  

  
  def _pk_to_address(self, public_key):
    """
    Given a pk object will return the simple text address.
    
    OBS: Should be overwritten in particular implementations using X962


    Parameters
    ----------
    public_key : pk
      the pk object.
      
    Returns
    -------
      text address      
    
    """
    data = public_key.public_bytes(
      encoding=serialization.Encoding.DER, # will encode the full pk information 
      format=serialization.PublicFormat.SubjectPublicKeyInfo, # used with DER
    )
    txt = BCct.ADDR_PREFIX + self._binary_to_text(data)
    return txt


  def _address_to_pk(self, address):
    """
    Given a address will return the pk object
    
    OBS: Should be overwritten in particular implementations using X962


    Parameters
    ----------
    address : str
      the text address (pk).


    Returns
    -------
    pk : pk
      the pk object.

    """
    simple_address = self._remove_prefix(address)
    bpublic_key = self._text_to_binary(simple_address)
    # below works for DER / SubjectPublicKeyInfo
    public_key = serialization.load_der_public_key(bpublic_key)
    return public_key
  
  
  def _text_to_sk(self, source, from_file=False, password=None):
    """
    Construct a EllipticCurvePrivateKey from a text sk

    Parameters
    ----------
    source : str
      the text secret key or the file name if `from_file == True`.
      
    from_file: bool
      flag that allows source to be a file
      

    Returns
    -------
      sk

    """
    if from_file and os.path.isfile(source):
      self.P("Reading SK from '{}'".format(source), verbosity=2)
      with open(source, 'rt') as fh:
        data = fh.read()
    else:
      data = source
    
    bdata = data.encode()
    if password:
      pass_data = password.encode()
    else:
      pass_data = None
    private_key = serialization.load_pem_private_key(bdata, pass_data)
    return private_key
  
  def _sk_to_text(self, private_key, password=None, fn=None):
    """
    Serialize a sk as text

    Parameters
    ----------
    private_key : sk
      the secret key object.
      
    password: str
      password to be used for sk serialization
      
    fn: str:
      text file where to save the pk

    Returns
    -------
      the sk as text string

    """
    if password:
      encryption_algorithm = serialization.BestAvailableEncryption(password.encode())
    else:
      encryption_algorithm = serialization.NoEncryption()
      
    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=encryption_algorithm        
    )     
    str_pem = pem.decode()
    if fn is not None:
      full_path = os.path.abspath(fn)
      self.P("Writing PEM-encoded key to {}".format(full_path), color='g', verbosity=2)
      with open(fn, 'wt') as fh:
        fh.write(str_pem)
    return str_pem  
  
  
  def _dict_to_json(self, dct_data, replace_nan=True, inplace=True):
    if replace_nan:
      dct_safe_data = replace_nan_inf(dct_data, inplace=inplace)
    else:
      dct_safe_data = dct_data
      
    dumps_config = dict(
      cls=_ComplexJsonEncoder, 
      separators=(',',':'),
      ensure_ascii=self.__ensure_ascii_payloads,
    )
    # we dump the data to a string then we reload and sort as there might
    # be some issues with the sorting if we have int keys that will be sorted
    # then recovered as string keys
    str_data = json.dumps(dct_safe_data, **dumps_config)
    # we reload the data to ensure 
    dct_reload = json.loads(str_data)
    # in order to ensure the data is sorted we dump it again to a string
    # IMPORTANT: we need to use the same dumps_config as before !
    str_data = json.dumps(dct_reload, sort_keys=True, **dumps_config) 
    return str_data
  
  def _create_new_sk(self):
    """
    Simple wrapper to generated pk


    Returns
    -------
    private_key : sk
    
    """
    raise NotImplementedError()
    

  
  def _verify(self, public_key, signature : bytes, data : bytes):
    """
    Verifies a `pk` signature on a binary `data` package
    

    Parameters
    ----------
    public_key : pk type
      the pk object.
      
    signature : bytes
      the binary signature.
      
    data : bytes
      the binary message.


    Returns
    -------
    result: _DotDict 
      contains `result.ok` and `result.message`

    """
    raise NotImplementedError()

  
  
  def _sign(self, data : bytes, private_key, text=False):
    """
    Sign a binary message with Elliptic Curve
    

    Parameters
    ----------
    data : bytes
      the binary message.
      
    private_key : pk
      the private key object.
      
    text : bool, optional
      return the signature as text. The default is False.

    Returns
    -------
    signature as text or binary

    """
    raise NotImplementedError()  
  
      
  
  #############################################################################
  ####                                                                     ####
  ####                          Public functions                           ####
  ####                                                                     ####
  #############################################################################
  
  
  def contains_current_address(self, lst_addresses):
    """
    Checks if the current address is in the list of addresses

    Parameters
    ----------
    lst_addresses : list
      the list of addresses.

    Returns
    -------
    bool
      True if the current address is in the list.

    """
    return self.address_in_list(self.address, lst_addresses)
  
  def address_in_list(self, node_address, lst_addresses):
    """
    Checks if the address is in the list of addresses

    Parameters
    ----------
    node_address : str
      the address.
      
    lst_addresses : list
      the list of addresses.

    Returns
    -------
    bool
      True if the address is in the list.

    """
    node_address = self._remove_prefix(node_address)
    lst = [self._remove_prefix(x) for x in lst_addresses if x is not None]
    return node_address in lst
  
  @property
  def address(self):
    """Returns the public address"""
    return self.__address
    
  
  @property
  def address_no_prefix(self):
    """Returns the public address without the prefix"""
    return self._remove_prefix(self.address)
  
  @property
  def allowed_list(self):
    """Returns the allowed command senders (non-prefixed) for the current node"""
    return self._load_and_maybe_create_allowed(return_names=False, return_prefix=False)
  
  @property
  def whitelist(self):
    """Returns the allowed command senders for the current node"""
    return self.allowed_list
  
  @property
  def whitelist_with_names(self):
    """
    Returns a tuple with the allowed list (prefixed) and a list with names
    """
    return self._load_and_maybe_create_allowed(return_names=True, return_prefix=True)
  
  
  @property
  def whitelist_with_prefixes(self):
    """
    Returns the allowed command senders (prefixed) for the current node
    """
    return self._load_and_maybe_create_allowed(return_names=False, return_prefix=True)

  
  def maybe_remove_prefix(self, address):
    """
    Removes the prefix from the address

    Parameters
    ----------
    address : str
      the text address.

    Returns
    -------
    address : str
      the address without the prefix.
    """
    return self._remove_prefix(address)  
  

  def maybe_add_prefix(self, address):
    """
    Adds the prefix to the address

    Parameters
    ----------
    address : str
      the text address.

    Returns
    -------
    address : str
      the address with the prefix.
    """
    return self._add_prefix(address)
    
  
  def dict_digest(self, dct_data, return_str=True):
    """Generates the hash of a dict object given as parameter"""
    str_data = self._dict_to_json(dct_data, replace_nan=True)
    bin_hex_hash, hex_hash = self._compute_hash(str_data.encode())
    if return_str:
      return hex_hash
    else:
      return bin_hex_hash
  
  
  def save_sk(self, fn, password=None):
    """
    Saves the SK with or without password

    Parameters
    ----------
    fn : str
      SK file name.
    password : str, optional
      optional password. The default is None.

    Returns
    -------
    fn : str
      saved file name.

    """
    self.P("Serializing the private key...", verbosity=2)
    _ = self._sk_to_text(
      private_key=self.__private_key,
      password=password,
      fn=fn
    )
    return fn
  
  
  def _generate_data_for_hash(self, dct_data, replace_nan=True):
    """
    Will convert the dict to json (removing the non-data fields) and return the json string. 
    The dict will be modified inplace to replace NaN and Inf with None.
    """
    assert isinstance(dct_data, dict), "Cannot compute hash on non-dict data"
    if self.eth_enabled:
      dct_only_data = {k:dct_data[k] for k in dct_data if k not in ALL_NON_DATA_FIELDS}
    else:
      dct_only_data = {k:dct_data[k] for k in dct_data if k not in NO_ETH_NON_DATA_FIELDS}
    #endif
    str_data = self._dict_to_json(
      dct_only_data, 
      replace_nan=replace_nan, 
      inplace=True # will replace inplace the np.nan and np.inf with None
    )
    return str_data
    
  
  def compute_hash(self, dct_data, return_all=False, replace_nan=True):
    """
    Computes the hash of a dict object

    Parameters
    ----------
    dct_data : dict
      the input message as a dict.
      
    return_all: bool, optional
      if `True` will return the binary hash as well. Default `False`
      
    replace_nan: bool, optional
      will replace inplace `np.nan` and `np.inf` with `None` before hashing. Default `True`

    Returns
    -------
    result : str or tuple(bytes, bytes, str) if `return_all` is `True`
      
    """
    str_data = self._generate_data_for_hash(dct_data, replace_nan=replace_nan)
    bdata = bytes(str_data, 'utf-8')
    bin_hexdigest, hexdigest = self._compute_hash(bdata)
    if return_all:
      result = bdata, bin_hexdigest, hexdigest
    else:
      result = hexdigest
    return result
  
  
  def sign(self, dct_data: dict, add_data=True, use_digest=True, replace_nan=True, eth_sign=False) -> str:
    """
    Generates the signature for a dict object.
    Does not add the signature to the dict object


    Parameters
    ----------
    dct_data : dict
      the input message as a dict.
      
    add_data: bool, optional
      will add signature and address to the data dict (also digest if required). Default `True`
      
    use_digest: bool, optional  
      will compute data hash and sign only on hash
      
    replace_nan: bool, optional
      will replace `np.nan` and `np.inf` with `None` before signing. 
      
    eth_sign: bool, optional
      will also sign the data with the Ethereum account. Default `False`

    Returns
    -------
      text signature

        
      IMPORTANT: 
        It is quite probable that the same sign(sk, hash) will generate different signatures

    """
    result = None
    assert isinstance(dct_data, dict), "Cannot sign on non-dict data"
    
    bdata, bin_hexdigest, hexdigest = self.compute_hash(
      dct_data, 
      return_all=True, 
      replace_nan=replace_nan,
    )
    text_data = bdata.decode()
    if use_digest:
      bdata = bin_hexdigest # to-sign data is the hash
    # finally sign either full or just hash
    result = self._sign(data=bdata, private_key=self.__private_key, text=True)
    if add_data:
      # now populate dict
      dct_data[BCct.SIGN] = result
      dct_data[BCct.SENDER] = self.address
      
      if self._eth_enabled:
        dct_data[BCct.ETH_SENDER] = self.eth_address
        ### add eth signature
        dct_data[BCct.ETH_SIGN] = "0xBEEF"
        if eth_sign:
          eth_sign_info = self.eth_sign_text(text_data, signature_only=False)
          # can be replaced with dct_data[BCct.ETH_SIGN] = self.eth_sign_text(bdata.decode(), signature_only=True)
          eth_sign = eth_sign_info.get('signature')
          dct_data[BCct.ETH_SIGN] = eth_sign
        ### end eth signature
      if use_digest:
        dct_data[BCct.HASH] = hexdigest
    return result
    
  
  
  def verify(
      self, 
      dct_data: dict, 
      signature: str=None, 
      sender_address: str=None, 
      return_full_info=True,
      verify_allowed=False,
      replace_nan=True,
      log_hash_sign_fails=True,
    ):
    """
    Verifies the signature validity of a given text message

    Parameters
    ----------
    dct_data : dict
      dict object that needs to be verified against the signature.
        
    signature : str, optional
      the text encoded signature. Extracted from dict if missing
      
    sender_address : str, optional
      the text encoded public key. Extracted from dict if missing
      
    return_full_info: bool, optional
      if `True` will return more than `True/False` for signature verification
      
    verify_allowed: bool, optional
      if true will also check if the address is allowed by calling `check_allowed`
      
    replace_nan: bool, optional
      will replace `np.nan` and `np.inf` with `None` before verifying. Default `True`
    
    log_hash_sign_fails: bool, optional
      if `True` will log the verification failures for hash and signature issues. Default `True`
    

    Returns
    -------
    bool / VerifyMessage
      returns `True` if signature verifies else `False`. 
      returns `VerifyMessage` structure if return_full_info (default `True`)

    """
    result = False
    
    bdata_json, bin_hexdigest, hexdigest = self.compute_hash(
      dct_data, 
      return_all=True,
      replace_nan=replace_nan,
    )

    if signature is None:
      signature = dct_data.get(BCct.SIGN)
    
    if sender_address is None:
      sender_address = dct_data.get(BCct.SENDER)          
    
    verify_msg = VerifyMessage()
    verify_msg.sender = sender_address
    
    received_digest = dct_data.get(BCct.HASH)
    if received_digest:
      # we need to verify hash and then verify signature on hash      
      if hexdigest != received_digest:
        verify_msg.message = "Corrupted digest!"
        verify_msg.valid = False
      #endif hash failed
      bdata = bin_hexdigest
    else:
      # normal signature on data
      bdata = bdata_json
    #endif has hash or not
    
    if verify_msg.message is None:            
      try:
        assert sender_address is not None, 'Sender address is NULL'
        assert signature is not None, 'Signature is NULL'
        
        bsignature = self._text_to_binary(signature)
        pk = self._address_to_pk(sender_address)
        verify_msg = self._verify(public_key=pk, signature=bsignature, data=bdata)
      except Exception as exc:
        verify_msg.message = str(exc)
        verify_msg.valid = False
    #endif check if signature failed already from digesting

    verify_msg.sender = sender_address
    
    if not verify_msg.valid:
      if log_hash_sign_fails and signature is not None and sender_address is not None:
        self.P("Signature failed on msg from {}: {}".format(
          sender_address, verify_msg.message
          ), color='r', verbosity=1,
        )
    elif verify_allowed and verify_msg.valid:
      if not self.is_allowed(sender_address):
        verify_msg.message = "Signature ok but address {} not in {}.".format(sender_address, BCct.AUTHORISED_ADDRS)
        verify_msg.valid = False
      #endif not allowed
    #endif ok but authorization required
    
    if return_full_info:
      result = verify_msg
    else:
      result = verify_msg.ok
    return result
  
  
  def is_allowed(self, sender_address: str):
    to_search_address = self._remove_prefix(sender_address)
    is_allowed = to_search_address in self.allowed_list or to_search_address == self._remove_prefix(self.address)
    return is_allowed
  
  
  def encrypt(self, data, destination):
    """
    Encrypts the data for a given destination

    Parameters
    ----------
    data : dict
      the data to be encrypted.
      
    destination : str
      the destination address.

    Returns
    -------
    None.

    """
    raise NotImplementedError()
      
  def decrypt(self, data):
    """
    Decrypts the data

    Parameters
    ----------
    data : dict
      the data to be decrypted.

    Returns
    -------
    None.

    """
    raise NotImplementedError()
  
  
  

  def dauth_autocomplete(
    self, 
    dauth_endp=None, 
    add_env=True, 
    debug=False, 
    max_tries=5,
    network=None,
    return_full_data=False,
    debug_data=False,
    **kwargs
  ):
    """
    Autocompletes the environment with the dAuth information.
    Parameters
    ----------
    dauth_endp
    add_env
    debug
    max_tries
    network
    kwargs

    Returns
    -------
    None if the URL is invalid or the request failed.
    dict with the dAuth information if the request got status 200(if errors occured, but
    the status is still 200, an empty dictionary will be returned).
    """
    if EE_VPN_IMPL:
      return None # must return None not empty dict for VPNs
    #endif EE_VPN_IMPL
    
    from ratio1._ver import __VER__ as sdk_version
    try:
      from ver import __VER__ as app_version
    except:
      app_version = None
    try:
      from naeural_core.main.ver import __VER__ as core_version
    except:
      core_version = None
      
    MIN_LEN = 10
    # Default result in case of invalid URL should be None
    # An empty dict will mean that the URL is valid, but the node
    # does not have a license associated with it.
    dct_env = None
    done = False
    tries = 0
    in_env = False
    url = dauth_endp
    dct_response = {}

    # Network handling
    if network is None:
      network = self.evm_network

    # URL handling
    if not isinstance(url, str) or len(url) < MIN_LEN:
      if DAUTH_ENV_KEY in os.environ:
        in_env = True
        url = os.environ[DAUTH_ENV_KEY]
      else:
        network_data = self.get_network_data(network)
        url = network_data[dAuth.EvmNetData.DAUTH_URL_KEY]
      #endif not in env
      
    if isinstance(url, str) and len(url) > MIN_LEN:
      # Valid URL
      if dauth_endp is None:
        if in_env:
          self.P("Found dAuth URL in environment: '{}'".format(url))
        else:
          self.P("Using default dAuth URL: '{}'".format(url))
      eth_short = self.eth_address[:6] + '...' + self.eth_address[-4:]
      while not done:
        self.P(f"<{eth_short}> ({network}) dAuth with `{url}`... (try {tries + 1} / {max_tries})")
        try:
          if debug_data:
            to_send = {
              DAUTH_NONCE : str(uuid.uuid4())[:8],
              dAuth.DAUTH_SENDER_APP_VER  : app_version,
              dAuth.DAUTH_SENDER_SDK_VER  : sdk_version,
              dAuth.DAUTH_SENDER_CORE_VER : core_version,
              **kwargs,
            }
          else:
            to_send = {
              **kwargs,
              DAUTH_NONCE : str(uuid.uuid4())[:8],
              dAuth.DAUTH_SENDER_APP_VER  : app_version,
              dAuth.DAUTH_SENDER_SDK_VER  : sdk_version,
              dAuth.DAUTH_SENDER_CORE_VER : core_version,
            }
          ######
          if len(kwargs) == 0:
            to_send[dAuth.DAUTH_SENDER_ALIAS] = dAuth.DAUTH_SENDER_ALIAS_DEFAULT
          ######
          self.sign(to_send)    
          json_to_send = {'body' : to_send}
          if debug:
            self.P(f"Sending to dAuth URL: {url}\n{json.dumps(json_to_send, indent=2)}")   
          response = requests.post(url, json=json_to_send)
          if response.status_code == 200:
            dct_response = response.json()
            dct_result = dct_response.get('result', {}) or {}
            dct_dauth = dct_result.get(DAUTH_SUBKEY, {}) or {}
            server_alias = dct_result.get(
              dAuth.DAUTH_SERVER_ALIAS, dAuth.DAUTH_ALIAS_UNK
            )
            server_addr = dct_result.get(
              BCctbase.SENDER, dAuth.DAUTH_ADDR_UNK
            )
            if debug:
              self.P(f"Response received from {server_alias} <{server_addr}>:\n {json.dumps(dct_response, indent=2)}")
            else:
              self.P(f"Response received from {server_alias} <{server_addr}>,")
            ver_result = self.verify(dct_result)
            if ver_result.valid:
              self.P(f"Signature from {server_alias} <{server_addr}> is valid.", color='g')
            else:
              self.P(f"Signature from {server_alias} <{server_addr}> is INVALID: {ver_result}", color='r')
              # Invalid response signature, thus return {}
              return {}

            # whitelist
            whitelist = dct_dauth.pop(dAuth.DAUTH_WHITELIST, [])
            if isinstance(whitelist, (str, list)) and len(whitelist) > 0:
              if isinstance(whitelist, str):
                whitelist = [whitelist]
              self.P(f"Found {len(whitelist)} whitelist addresses in dAuth response.", color='y')
              self.add_address_to_allowed(whitelist)
            else:
              self.P(f"No whitelist addresses found in dAuth response.", color='d')
            # end whitelist

            error = dct_dauth.get('error', None)

            dct_env = {k : v for k,v in dct_dauth.items() if k.startswith(dAuth.DAUTH_ENV_KEYS_PREFIX)}
            if len(dct_env) > 0:
              self.P("Found {} keys in dAuth response.".format(len(dct_env)), color='g')
              for k, v in dct_env.items():
                try:
                  if not isinstance(v, str):
                    v = json.dumps(v)
                except:
                  v = str(v)
                if k not in os.environ:
                  self.P(f"  Adding key `{k}{'=' + str(v) + ' ({})'.format(type(v).__name__) if debug else ''}` to env.", color='y')
                else:
                  self.P(f"  Overwrite  `{k}{'=' + str(v) + ' ({})'.format(type(v).__name__) if debug else ''}` in env.", color='y')
                if add_env:
                  os.environ[k] = v
                #endif add to env
              #endfor each key in dAuth response
              if error is not None:
                self.P(f"Server message: {error}", color='y')
            else:
              self.P(f"dAuth reject response. Your node is not licensed. Error: {error}", color='r')
            done = True
          else:
            self.P(f"Error in dAuth response: {response.status_code} - {response.text}", color='r')
        except Exception as exc:
          self.P(f"Error in dAuth URL request: {exc}. Received: {dct_response}", color='r')
          # Request failed somewhere so dct_env will be set again to None.
          dct_env = None
        #end try
        tries += 1
        if tries >= max_tries:
          done = True    
      #end while
    else:
      # Invalid URL, thus dct_env will remain None
      self.P(f"dAuth URL is not valid: {url}", color='r')
    #end if url is valid
    if return_full_data:
      return dct_response
    return dct_env



"""
cli.py
======

This script serves as the command-line interface (CLI) for the Ratio1 SDK.
It provides a user-friendly way to interact with the SDK's features and
functionalities through various commands and options.
"""

import signal
import sys
import argparse

from ratio1.utils.config import maybe_init_config, log_with_color
from ratio1.cli.cli_commands import CLI_COMMANDS
from ratio1 import version
import traceback

def handle_sigint(signum, frame):
  """Handler for the SIGINT signal (Ctrl-C)."""
  print("Interrupted. Exiting...")
  sys.exit(1)

def create_global_parser():
  """
  Creates a global argument parser with shared options like verbosity.

  Returns
  -------
  argparse.ArgumentParser
      Global argument parser.
  """
  global_parser = argparse.ArgumentParser(add_help=False)
  global_parser.add_argument(
    "-v", "--verbose", action="store_true", help="Enable verbose output"
  )
  return global_parser

def build_parser():
  """
  Dynamically builds the argument parser based on CLI_COMMANDS.

  Returns
  -------
  argparse.ArgumentParser
      Configured argument parser.
  """
  global_parser = create_global_parser()

  title = f"Ratio1 fleet control for SDK v{version} - CLI for ratio1 Edge Protocol SDK package"
  parser = argparse.ArgumentParser(description=title, parents=[global_parser])

  # Top-level subparsers (e.g. 'get', 'config', 'restart', 'shutdown')
  subparsers = parser.add_subparsers(dest="command", help="Available commands")
  subparsers.required = True  # Force user to pick a valid command

  for command, subcommands in CLI_COMMANDS.items():
    # Create a sub-parser for each top-level command
    cmd_help = subcommands.get("description") if "func" in subcommands else f"{command} commands"
    command_parser = subparsers.add_parser(command, help=cmd_help)

    # Check if this command has nested subcommands vs. a single-level function
    if isinstance(subcommands, dict) and "func" not in subcommands:
      # Nested subcommands (e.g. 'get nodes', 'get apps', etc.)
      command_subparsers = command_parser.add_subparsers(
        dest="subcommand",
        required=True,
        help=f"Subcommands under '{command}'"
      )

      for subcommand, subcmd_info in subcommands.items():
        description = subcmd_info.get("description", f"{subcommand} command")
        subcommand_parser = command_subparsers.add_parser(subcommand, help=description)

        # If we have parameters, add them
        if isinstance(subcmd_info, dict) and "params" in subcmd_info:
          for param, desc in subcmd_info["params"].items():
            if param.startswith("--"):
              if desc.lower().endswith("(flag)"):
                subcommand_parser.add_argument(param, action="store_true", help=desc)
              else:
                subcommand_parser.add_argument(param, type=str, help=desc)
            else:
              # Positional argument
              subcommand_parser.add_argument(param, help=desc)

        # Assign the function to call
        subcommand_parser.set_defaults(func=subcmd_info["func"])
    else:
      # Single-level command with optional parameters
      if "params" in subcommands:
        for param, desc in subcommands["params"].items():
          if param.startswith("--"):
            if desc.lower().endswith("(flag)"):
              command_parser.add_argument(param, action="store_true", help=desc)
            else:
              command_parser.add_argument(param, type=str, help=desc)
          else:
            command_parser.add_argument(param, help=desc)

      # Set the function to be called for this command
      command_parser.set_defaults(func=subcommands["func"])

  return parser

def main():
  """
  Main entry point for the CLI.
  Ensures the configuration is initialized, builds the parser,
  and executes the appropriate command function.
  """
  # Register SIGINT handler
  signal.signal(signal.SIGINT, handle_sigint)
  print("Processing...\r", end="", flush=True)

  try:
    initialized = maybe_init_config()

    if initialized:
      parser = build_parser()
      args = parser.parse_args()

      if hasattr(args, "func"):
        args.func(args)
      else:
        parser.print_help()

  except Exception as e:
    log_with_color(f"Error: {e}:\n{traceback.format_exc()}", color='r')

if __name__ == "__main__":
  main()



from . import base as BASE_CT
# BC Consts:
from ..bc.base import BCct as BC_CT

SB_ID = BASE_CT.SB_ID
EE_ID = BASE_CT.EE_ID
EE_ADDR = 'EE_ADDR'

HOST = 'HOST'
TOPIC = 'TOPIC'
BROKER = 'BROKER'
PORT = 'PORT'
USER = 'USER'
PASS = 'PASS'
CERT_PATH = 'CERT_PATH'
VHOST = 'VHOST'
DEVICE_ID = 'DEVICE_ID'
ROUTING_KEY = 'ROUTING_KEY'
QOS = 'QOS'
SECURED = 'SECURED'

EXCHANGE = 'EXCHANGE'
EXCHANGE_TYPE = 'EXCHANGE_TYPE'
QUEUE = 'QUEUE'
QUEUE_DURABLE = 'QUEUE_DURABLE'
QUEUE_EXCLUSIVE = 'QUEUE_EXCLUSIVE'
QUEUE_DEVICE_SPECIFIC = 'QUEUE_DEVICE_SPECIFIC'

COMMUNICATION_CONFIG_CHANNEL = 'CONFIG_CHANNEL'
COMMUNICATION_PAYLOADS_CHANNEL = 'PAYLOADS_CHANNEL'
COMMUNICATION_CTRL_CHANNEL = 'CTRL_CHANNEL'
COMMUNICATION_NOTIF_CHANNEL = 'NOTIF_CHANNEL'
COMMUNICATION_VALID_CHANNELS = [
    COMMUNICATION_CONFIG_CHANNEL, COMMUNICATION_PAYLOADS_CHANNEL,
    COMMUNICATION_CTRL_CHANNEL, COMMUNICATION_NOTIF_CHANNEL
]
COMMUNICATION_DEFAULT = 'DEFAULT'
COMMUNICATION_COMMAND_AND_CONTROL = 'COMMANDCONTROL'
COMMUNICATION_HEARTBEATS = 'HEARTBEATS'
COMMUNICATION_NOTIFICATIONS = 'NOTIFICATIONS'
COMMUNICATION_VALID_TYPES = [
    COMMUNICATION_DEFAULT, COMMUNICATION_COMMAND_AND_CONTROL,
    COMMUNICATION_HEARTBEATS, COMMUNICATION_NOTIFICATIONS
]


class TIMERS:
  TIMER_SEND_BUFFER_PAYLOAD = 'send_buffer_payload'


class COMM_SEND_MESSAGE:
  K_ACTION = 'ACTION'
  K_PAYLOAD = 'PAYLOAD'
  K_EE_ID = EE_ID
  K_INITIATOR_ID = 'INITIATOR_ID'
  K_SESSION_ID = 'SESSION_ID'
  K_SENDER_ADDR = BC_CT.SENDER
  K_TIME = 'TIME'
  K_EE_IS_ENCRYPTED = "EE_IS_ENCRYPTED"
  K_EE_ENCRYPTED_DATA = "EE_ENCRYPTED_DATA"

  ACTION_VALUE_PAYLOAD = K_PAYLOAD


class COMM_RECV_MESSAGE:
  K_ACTION = COMM_SEND_MESSAGE.K_ACTION
  K_PAYLOAD = COMM_SEND_MESSAGE.K_PAYLOAD
  K_EE_ID = COMM_SEND_MESSAGE.K_EE_ID
  K_INITIATOR_ID = COMM_SEND_MESSAGE.K_INITIATOR_ID
  K_SESSION_ID = COMM_SEND_MESSAGE.K_SESSION_ID
  K_SENDER_ADDR = COMM_SEND_MESSAGE.K_SENDER_ADDR
  K_EE_IS_ENCRYPTED = COMM_SEND_MESSAGE.K_EE_IS_ENCRYPTED
  K_EE_ENCRYPTED_DATA = COMM_SEND_MESSAGE.K_EE_ENCRYPTED_DATA

  K_VALIDATED = 'VALIDATED'  # this is a flag to indicate if the message was validated or not


COMM_SEND_BUFFER = 100
COMM_RECV_BUFFER = 100
COMM_SECS_SHOW_INFO = 180



result = None
if plugin.int_cache['run_first_time'] == 0:
  # this is the first run, consider this the setup

  plugin.int_cache['run_first_time'] = 1

  worker_code = plugin.cfg_worker_code
  n_workers = plugin.cfg_n_workers
  # we use DeAPI `plugin.deapi_get_wokers` call to get the needed workers
  plugin.obj_cache['lst_workers'] = plugin.deapi_get_wokers(n_workers)
  plugin.obj_cache['dct_workers'] = {}
  plugin.obj_cache['dct_worker_progress'] = {}
  plugin.P(plugin.obj_cache['lst_workers'])

  # for each worker we symetrically launch the same job
  for worker in plugin.obj_cache['lst_workers']:
    plugin.obj_cache['dct_worker_progress'][worker] = []
    pipeline_name = plugin.cmdapi_start_simple_custom_pipeline(
      base64code=worker_code,
      dest=worker,
      instance_config={
        'MAX_TRIES': plugin.cfg_max_tries,
      }
    )
    plugin.obj_cache['dct_workers'][worker] = pipeline_name
  # endfor

  plugin.obj_cache["start_time"] = plugin.datetime.now()
  # endfor
elif (plugin.datetime.now() - plugin.obj_cache["start_time"]).seconds > plugin.cfg_max_run_time:
  # if the configured time has elapsed we stop all the worker pipelines
  # as well as stop this pipeline itself

  for node_id, pipeline_name in plugin.obj_cache['dct_workers'].items():
    plugin.cmdapi_archive_pipeline(dest=node_id, name=pipeline_name)
  # now archive own pipeline
  plugin.cmdapi_archive_pipeline()
  result = {
    'STATUS': 'DONE',
    'RESULTS': plugin.obj_cache['dct_worker_progress']
  }
else:
  # here are the operations we are running periodically
  payload = plugin.dataapi_struct_data()  # we use the DataAPI to get upstream data
  if payload is not None:

    node_id = payload.get('EE_ID', payload.get('SB_ID'))
    pipeline_name = payload.get('STREAM_NAME')

    if (node_id, pipeline_name) in plugin.obj_cache['dct_workers'].items():
      # now we extract result from the result key of the payload JSON
      # this also can be configured to another name
      num = payload.get('EXEC_RESULT', payload.get('EXEC_INFO'))
      if num is not None:
        plugin.obj_cache['dct_worker_progress'][node_id].append(num)
        result = {
          'STATUS': 'IN_PROGRESS',
          'RESULTS': plugin.obj_cache['dct_worker_progress']
        }
  # endif
# endif



"""
ex13_launch_repo_based_webapp.py
-------------------------------

This script demonstrates how to create and deploy a webapp based on a public repository using ngrok
edge definition. The script will create a webapp based on a public repository and deploy it to the 
target Ratio1 Edge Node and assing it the same ngrok edge label. 

TODO:
This will be modified so that it will start a generic webapp and generates a dynamic URL that can 
be used to access the webapp. For this demo, we will use a custom webapp that resides at a 
public repository location.

"""
from os import environ
from ratio1 import Session, PLUGIN_SIGNATURES


if __name__ == "__main__":
  session = Session()

  nodes = [
    environ.get("EE_TARGET_NODE_1", "INSERT_YOUR_NODE_ADDRESS_HERE"),
    environ.get("EE_TARGET_NODE_2", "INSERT_YOUR_NODE_ADDRESS_HERE")
  ]

  if isinstance(nodes, str):
    nodes = [nodes]

  for node in nodes:
    session.wait_for_node(node)

  # In case this is not provided and create_and_deploy_balanced_web_app is used an exception will be raised.
  # In case this is not provided by the user, but create_webapp method is used, a unique URL will be generated by the ngrok service.
  # Warning! create_webapp method does not support multiple nodes.
  ngrok_edge_label = environ.get("NGROK_EDGE_LABEL", None)

  # Defining the environment variables for the webapp.
  port = 3000
  env_vars = {
    'APP_PORT': port,
    'LOCAL_ADDRESS': '/edge_node/_local_cache/_data/local_info.json'
  }

  # Defining the webapp setup commands.
  setup_commands = [
    "npm install"
  ]

  # Defining the assets of the application.
  assets = {
    'operation': "clone",
    "url": "https://github.com/ratio1/demo-deploy-nodejs",
  }

  # Defining the webapp run commands.
  run_command = "npm start"

  # instance: PLUGIN_TYPES.CUSTOM_WEBAPI_01
  session.create_and_deploy_balanced_web_app(
    nodes=nodes,
    name="Ratio1_WebApp_tutorial",
    signature=PLUGIN_SIGNATURES.GENERIC_WEB_APP,
    env_vars=env_vars,
    port=port,
    setup_commands=setup_commands,
    run_command=run_command,
    ngrok_edge_label=ngrok_edge_label,
    assets=assets
  )

  # Observation:
  #   next code is not mandatory - it is used to keep the session open and cleanup the resources
  #   in production, you would not need this code as the script can close after the pipeline will be sent
  session.run(
    wait=True,  # wait for the user to stop the execution
    close_pipelines=True  # when the user stops the execution, the remote edge-node pipelines will be closed
  )



from ...base import Instance, Pipeline
from ...const import PLUGIN_SIGNATURES

class CustomWebapi01(Instance):
  signature = PLUGIN_SIGNATURES.CUSTOM_WEBAPI_01

  def get_proposed_endpoints(self):
    from copy import deepcopy
    proposed_config = self._get_proposed_config_dictionary(full=True)
    if "ENDPOINTS" in proposed_config:
      return deepcopy(proposed_config["ENDPOINTS"])
    return deepcopy(self.config.get("ENDPOINTS", []))

  def get_endpoint_fields(self, method: callable):
    """
    A aliast for get_method_data: it returns the name, args and base64 code of a method.
    """  
    return self._get_method_data(method)


  def get_proposed_assets(self):
    from copy import deepcopy
    proposed_config = self._get_proposed_config_dictionary(full=True)
    if "ASSETS" in proposed_config:
      return deepcopy(proposed_config["ASSETS"])
    return deepcopy(self.config.get("ASSETS", {}))
  

  def get_proposed_jinja_args(self):
    from copy import deepcopy
    proposed_config = self._get_proposed_config_dictionary(full=True)
    if "JINJA_ARGS" in proposed_config:
      return deepcopy(proposed_config["JINJA_ARGS"])
    return deepcopy(self.config.get("JINJA_ARGS", {}))


  def add_new_endpoint(self, endpoint_type="default", **kwargs):
    """
    Add a new endpoint to a existing web app instance.

    Parameters
    ----------
    endpoint_type : str, optional
        The type of the endpoint. Can be "default", "file" or "html". The default is "default".

    Raises
    ------
    ValueError
        If the endpoint_type is invalid.
    """
    self.P("Attempting to add a new `{}` endpoint: {}".format(endpoint_type, kwargs))
    if endpoint_type == "default":
      self.add_new_function_endpoint(**kwargs)
    elif endpoint_type == "file":
      self.add_new_file_endpoint(**kwargs)
    elif endpoint_type == "html":
      self.add_new_html_endpoint(**kwargs)
    else:
      raise ValueError("Invalid endpoint type.")
    return


  def add_new_file_endpoint(self, str_code, file_name, endpoint_name):
    raise NotImplementedError("This method is not implemented yet.")
  
  
  def add_new_function_endpoint(self, function, method="get"):
    name, args, base64_code = self.get_endpoint_fields(function)
    dct_endpoint = {
      "NAME": name
    }

    proposed_endpoints = self.get_proposed_endpoints()
    lst_pos = [pos for pos, endpoint in enumerate(proposed_endpoints) if endpoint["NAME"] == name]

    if len(lst_pos) > 0:
      dct_endpoint = proposed_endpoints[lst_pos[0]]
    else:
      proposed_endpoints.append(dct_endpoint)
    # endif

    dct_endpoint["CODE"] = base64_code
    dct_endpoint["METHOD"] = method
    dct_endpoint["ARGS"] = args

    self.update_instance_config(config={"ENDPOINTS": proposed_endpoints})
    return  


  def add_new_html_endpoint(self, html_path, web_app_file_name, endpoint_route):
    str_code = None
    with open(html_path, "r") as file:
      str_code = file.read()

    if str_code is None:
      raise ValueError(f"Could not read the file {html_path}")
    self.pipeline: Pipeline
    base64_html = self.pipeline.str_to_base64(str_code, compress=True)

    proposed_assets = self.get_proposed_assets()
    proposed_jinja_args = self.get_proposed_jinja_args()

    if isinstance(proposed_assets, str):
      proposed_assets = {
        "url": [],
        "operation": "decode",
      }
    elif isinstance(proposed_assets, dict):
      if proposed_assets.get("operation") != "decode":
        proposed_assets["operation"] = "decode"
        proposed_assets["url"] = []

    lst_pos = [pos
               for pos, code_name_pair
               in enumerate(proposed_assets["url"])
               if code_name_pair[1] == 'assets/' + web_app_file_name]

    if len(lst_pos) > 0:
      proposed_assets["url"][lst_pos[0]][0] = base64_html
    else:
      proposed_assets["url"].append([base64_html, 'assets/' + web_app_file_name])
    # endif

    if proposed_jinja_args is None:
      proposed_jinja_args = {
        'html_files': [],
      }
    elif proposed_jinja_args.get('html_files') is None:
      proposed_jinja_args['html_files'] = []

    lst_pos = [pos
               for pos, dict_name_route_method,
               in enumerate(proposed_jinja_args['html_files'])
               if dict_name_route_method['name'] == web_app_file_name]

    dict_name_route_method = {
      "name": web_app_file_name,
      "method": "get"
    }
    if len(lst_pos) > 0:
      dict_name_route_method = proposed_jinja_args['html_files'][lst_pos[0]]
    else:
      proposed_jinja_args['html_files'].append(dict_name_route_method)

    dict_name_route_method["route"] = endpoint_route

    self.update_instance_config(config={"ASSETS": proposed_assets, "JINJA_ARGS": proposed_jinja_args})
    return
